import astrodynx as adx
import jax.numpy as jnp


class TestMeanAnomalyElps:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        a = 1.0
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / a**3) * deltat
        result = adx.mean_anomaly_elps(a, mu, deltat)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        a = jnp.array([1.0, 2.0])
        mu = jnp.array([1.0, 2.0])
        deltat = jnp.array([1.0, 1.0])
        expected = jnp.sqrt(mu / a**3) * deltat
        result = adx.mean_anomaly_elps(a, mu, deltat)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        a = jnp.array([1.0, 2.0])
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / a**3) * deltat
        result = adx.mean_anomaly_elps(a, mu, deltat)
        assert jnp.allclose(result, expected)

    def test_types(self) -> None:
        """Test that the function returns the correct type."""
        a = 1.0
        mu = 1.0
        deltat = 1.0
        result = adx.mean_anomaly_elps(a, mu, deltat)
        assert isinstance(result, jnp.ndarray)


class TestMeanAnomalyHypb:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        a = -1.0
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / -(a**3)) * deltat
        result = adx.mean_anomaly_hypb(a, mu, deltat)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        a = jnp.array([-1.0, -2.0])
        mu = jnp.array([1.0, 2.0])
        deltat = jnp.array([1.0, 1.0])
        expected = jnp.sqrt(mu / -(a**3)) * deltat
        result = adx.mean_anomaly_hypb(a, mu, deltat)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        a = jnp.array([-1.0, -2.0])
        mu = 1.0
        deltat = 1.0
        expected = jnp.sqrt(mu / -(a**3)) * deltat
        result = adx.mean_anomaly_hypb(a, mu, deltat)
        assert jnp.allclose(result, expected)

    def test_types(self) -> None:
        """Test that the function returns the correct type."""
        a = -1.0
        mu = 1.0
        deltat = 1.0
        result = adx.mean_anomaly_hypb(a, mu, deltat)
        assert isinstance(result, jnp.ndarray)


class TestKeplerequElps:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        E = 1.0
        e = 0.5
        a = 2.0
        mu = 1.0
        expected = (E - e * jnp.sin(E)) * jnp.sqrt(a**3 / mu)
        result = adx.keplerequ_elps(E, e, a, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        E = jnp.array([1.0, 2.0])
        e = jnp.array([0.5, 0.3])
        a = jnp.array([2.0, 3.0])
        mu = jnp.array([1.0, 2.0])
        expected = (E - e * jnp.sin(E)) * jnp.sqrt(a**3 / mu)
        result = adx.keplerequ_elps(E, e, a, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        E = jnp.array([1.0, 2.0])
        e = 0.5
        a = 2.0
        mu = 1.0
        expected = (E - e * jnp.sin(E)) * jnp.sqrt(a**3 / mu)
        result = adx.keplerequ_elps(E, e, a, mu)
        assert jnp.allclose(result, expected)

    def test_types(self) -> None:
        """Test that the function returns the correct type."""
        E = 1.0
        e = 0.5
        a = 2.0
        mu = 1.0
        result = adx.keplerequ_elps(E, e, a, mu)
        assert isinstance(result, jnp.ndarray)


class TestKeplerequHypb:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        H = 1.0
        e = 1.5
        a = -2.0
        mu = 1.0
        expected = (e * jnp.sinh(H) - H) * jnp.sqrt(-(a**3) / mu)
        result = adx.keplerequ_hypb(H, e, a, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        H = jnp.array([1.0, 2.0])
        e = jnp.array([1.5, 1.3])
        a = jnp.array([-2.0, -3.0])
        mu = jnp.array([1.0, 2.0])
        expected = (e * jnp.sinh(H) - H) * jnp.sqrt(-(a**3) / mu)
        result = adx.keplerequ_hypb(H, e, a, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        H = jnp.array([1.0, 2.0])
        e = 1.5
        a = -2.0
        mu = 1.0
        expected = (e * jnp.sinh(H) - H) * jnp.sqrt(-(a**3) / mu)
        result = adx.keplerequ_hypb(H, e, a, mu)
        assert jnp.allclose(result, expected)

    def test_types(self) -> None:
        """Test that the function returns the correct type."""
        H = 1.0
        e = 1.5
        a = -2.0
        mu = 1.0
        result = adx.keplerequ_hypb(H, e, a, mu)
        assert isinstance(result, jnp.ndarray)


class TestKeplerequUni:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        chi = 1.0
        r0 = jnp.array([1.0, 0.0, 0.0])
        v0 = jnp.array([0.0, 1.0, 0.0])
        mu = 1.0
        expected = 1.0  # Value from the function's implementation
        result = adx.keplerequ_uni(chi, r0, v0, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        chi = jnp.array([1.0, 1.0])
        r0 = jnp.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        v0 = jnp.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0]])
        mu = jnp.array([[1.0, 1.0]])
        # Expected values calculated from the function's formula
        expected = jnp.array([1.0, 1.0])
        result = adx.keplerequ_uni(chi, r0, v0, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        chi = jnp.array([1.0, 2.0])
        r0 = jnp.array([1.0, 0.0, 0.0])
        v0 = jnp.array([0.0, 1.0, 0.0])
        mu = 1.0
        # Expected values for broadcasting
        expected = jnp.array([1.0, 2.0])
        result = adx.keplerequ_uni(chi, r0, v0, mu)
        assert jnp.allclose(result, expected)

    def test_types(self) -> None:
        """Test that the function returns the correct type."""
        chi = 1.0
        r0 = jnp.array([1.0, 0.0, 0.0])
        v0 = jnp.array([0.0, 1.0, 0.0])
        mu = 1.0
        result = adx.keplerequ_uni(chi, r0, v0, mu)
        assert isinstance(result, jnp.ndarray)



import jax.numpy as jnp
from jax.typing import ArrayLike
from jax import Array
from astrodynx.twobody.ivp import U1, U2, U3, sigma_func
import astrodynx as adx


def keplerequ_elps(E: ArrayLike, e: ArrayLike, a: ArrayLike, mu: ArrayLike) -> Array:
    r"""Returns the flight time from periapsis for an elliptical orbit.

    Args:
        E: Eccentric anomaly of the orbit.
        e: Eccentricity of the orbit; $e < 1$.
        a: Semimajor axis of the orbit, a > 0.
        mu: Gravitational parameter of the central body.

    Returns:
        The flight time from periapsis for an elliptical orbit.

    Notes:
        The flight time from periapsis for an elliptical orbit is calculated using the formula:
        $$
        \Delta t = (E - e \sin E) \sqrt{\frac{a^3}{\mu}}
        $$
        where $\Delta t$ is the flight time, $E$ is the eccentric anomaly, $e$ is the eccentricity, $a$ is the semimajor axis, and $\mu$ is the gravitational parameter.

    References:
        Battin, 1999, pp.160.

    Examples:
        A simple example of calculating the flight time for an orbit with eccentricity 0.1, eccentric anomaly π/4, semimajor axis 1.0, and gravitational parameter 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> e = 0.1
        >>> E = jnp.pi / 4
        >>> a = 1.0
        >>> mu = 1.0
        >>> adx.keplerequ_elps(E, e, a, mu)
        Array(0.7146875, dtype=float32, weak_type=True)

        With broadcasting, you can calculate the flight time for multiple eccentricities, eccentric anomalies, semimajor axes, and gravitational parameters:

        >>> e = jnp.array([0.1, 0.2])
        >>> E = jnp.array([jnp.pi / 4, jnp.pi / 3])
        >>> a = jnp.array([1.0, 2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> adx.keplerequ_elps(E, e, a, mu)
        Array([0.7146875, 1.747985 ], dtype=float32)
    """
    return (E - e * jnp.sin(E)) * jnp.sqrt(a**3 / mu)


def keplerequ_hypb(H: ArrayLike, e: ArrayLike, a: ArrayLike, mu: ArrayLike) -> Array:
    r"""Returns the flight time from periapsis for a hyperbolic orbit.

    Args:
        H: Hyperbolic eccentric anomaly of the orbit.
        e: Eccentricity of the orbit; $e > 1$.
        a: Semimajor axis of the orbit, a < 0.
        mu: Gravitational parameter of the central body.

    Returns:
        The flight time from periapsis for a hyperbolic orbit.

    Notes:
        The flight time from periapsis for a hyperbolic orbit is calculated using the formula:
        $$
        \Delta t = (e \sinh H - H) \sqrt{\frac{-a^3}{\mu}}
        $$
        where $\Delta t$ is the flight time, $H$ is the hyperbolic eccentric anomaly, $e > 1$ is the eccentricity, $a < 0$ is the semimajor axis, and $\mu$ is the gravitational parameter.

    References:
        Battin, 1999, pp.166.

    Examples:
        A simple example of calculating the flight time for an orbit with eccentricity 1.1, hyperbolic eccentric anomaly 1.0, semimajor axis -1.0, and gravitational parameter 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> e = 1.1
        >>> H = 1.0
        >>> a = -1.0
        >>> mu = 1.0
        >>> adx.keplerequ_hypb(H, e, a, mu)
        Array(0.29272127, dtype=float32, weak_type=True)

        With broadcasting, you can calculate the flight time for multiple eccentricities, hyperbolic eccentric anomalies, semimajor axes, and gravitational parameters:

        >>> e = jnp.array([1.1, 1.2])
        >>> H = jnp.array([1.0, 1.0])
        >>> a = jnp.array([-1.0, -2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> adx.keplerequ_hypb(H, e, a, mu)
        Array([0.29272127, 0.82048297], dtype=float32)
    """
    return (e * jnp.sinh(H) - H) * jnp.sqrt(-(a**3) / mu)


def keplerequ_uni(chi: ArrayLike, r0: ArrayLike, v0: ArrayLike, mu: ArrayLike) -> Array:
    r"""Returns the flight time from $t_0$ to the moment when the generalized anomaly is $\chi$.

    Args:
        chi: The generalized anomaly.
        r0: (..., 3) The position vector at $t_0$.
        v0: (..., 3) The velocity vector at $t_0$.
        mu: The gravitational parameter.

    Returns:
        The flight time from $t_0$ to the moment when the generalized anomaly is $\chi$.

    Notes:
        The flight time from $t_0$ to the moment when the generalized anomaly is $\chi$ is calculated using the formula:
        $$
        \Delta t = \frac{r_0 U_1(\chi, \alpha) + \sigma_0 U_2(\chi, \alpha) + U_3(\chi, \alpha)}{\sqrt{\mu}}
        $$
        where $\Delta t = t - t_0$ is the flight time, $r_0$ is the norm of the position vector at $t_0$, $v_0$ is the norm of the velocity vector at $t_0$, $\mu$ is the gravitational parameter, $U_1$ is the universal function U1, $U_2$ is the universal function U2, $U_3$ is the universal function U3, and $\sigma_0$ is the sigma function at $t_0$.

    References:
        Battin, 1999, pp.178.

    Examples:
        A simple example of calculating the flight time for an orbit with a generalized anomaly of 1.0, a position vector of [1, 0, 0], a velocity vector of [0, 1, 0], and a gravitational parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> r0 = jnp.array([1.0, 0.0, 0.0])
        >>> v0 = jnp.array([0.0, 1.0, 0.0])
        >>> mu = 1.0
        >>> adx.keplerequ_uni(chi, r0, v0, mu)
        Array([1.], dtype=float32)

        With broadcasting, you can calculate the flight time for multiple generalized anomalies, position and velocity vectors, and gravitational parameters:

        >>> chi = jnp.array([[1.0], [1.0]])
        >>> r0 = jnp.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        >>> v0 = jnp.array([[0.0, 1.0, 0.0], [-1.0, 0.0, 0.0]])
        >>> mu = jnp.array([[1.0], [1.0]])
        >>> adx.keplerequ_uni(chi, r0, v0, mu)
        Array([[1.],
               [1.]], dtype=float32)
    """
    r0_norm = jnp.linalg.norm(r0, axis=-1, keepdims=True)
    alpha = 1.0 / adx.semimajor_axis(
        r0_norm, jnp.linalg.norm(v0, axis=-1, keepdims=True), mu
    )
    return (
        r0_norm * U1(chi, alpha)
        + sigma_func(r0, v0, mu) * U2(chi, alpha)
        + U3(chi, alpha)
    ) / jnp.sqrt(mu)


def mean_anomaly_elps(a: ArrayLike, mu: ArrayLike, deltat: ArrayLike) -> Array:
    r"""Returns the mean anomaly for an elliptical orbit.

    Args:
        a: Semimajor axis of the orbit, a > 0.
        mu: Gravitational parameter of the central body.
        deltat: Time since periapsis passage.

    Returns:
        The mean anomaly for an elliptical orbit.

    Notes:
        The mean anomaly for an elliptical orbit is calculated using the formula:
        $$
        M = \sqrt{\frac{\mu}{a^3}} \Delta t
        $$
        where $M$ is the mean anomaly, $a>0$ is the semimajor axis, $\mu$ is the gravitational parameter, and $\Delta t$ is the time since periapsis passage.

    References:
        Battin, 1999, pp.160.

    Examples:
        A simple example of calculating the mean anomaly for an orbit with semimajor axis 1.0, gravitational parameter 1.0, and time since periapsis passage 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> a = 1.0
        >>> mu = 1.0
        >>> deltat = 1.0
        >>> adx.mean_anomaly_elps(a, mu, deltat)
        Array(1., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the mean anomaly for multiple semimajor axes, gravitational parameters, and times since periapsis passage:

        >>> a = jnp.array([1.0, 2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> deltat = jnp.array([1.0, 1.0])
        >>> adx.mean_anomaly_elps(a, mu, deltat)
        Array([1. , 0.5], dtype=float32)
    """
    return jnp.sqrt(mu / a**3) * deltat


def mean_anomaly_hypb(a: ArrayLike, mu: ArrayLike, deltat: ArrayLike) -> Array:
    r"""Returns the mean anomaly for a hyperbolic orbit.

    Args:
        a: Semimajor axis of the orbit, a < 0.
        mu: Gravitational parameter of the central body.
        deltat: Time since periapsis passage.

    Returns:
        The mean anomaly for a hyperbolic orbit.

    Notes:
        The mean anomaly for a hyperbolic orbit is calculated using the formula:
        $$
        N = \sqrt{\frac{\mu}{-a^3}} \Delta t
        $$
        where $N$ is the mean anomaly, $a<0$ is the semimajor axis, $\mu$ is the gravitational parameter, and $\Delta t$ is the time since periapsis passage.

    References:
        Battin, 1999, pp.166.

    Examples:
        A simple example of calculating the mean anomaly for an orbit with semimajor axis -1.0, gravitational parameter 1.0, and time since periapsis passage 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> a = -1.0
        >>> mu = 1.0
        >>> deltat = 1.0
        >>> adx.mean_anomaly_hypb(a, mu, deltat)
        Array(1., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the mean anomaly for multiple semimajor axes, gravitational parameters, and times since periapsis passage:

        >>> a = jnp.array([-1.0, -2.0])
        >>> mu = jnp.array([1.0, 2.0])
        >>> deltat = jnp.array([1.0, 1.0])
        >>> adx.mean_anomaly_hypb(a, mu, deltat)
        Array([1. , 0.5], dtype=float32)
    """
    return jnp.sqrt(mu / -(a**3)) * deltat



import astrodynx as adx
import jax.numpy as jnp


class TestSigmaFunc:
    def test_scalar_inputs(self) -> None:
        """Test with scalar inputs."""
        r = jnp.array([1.0, 0.0, 0.0])
        v = jnp.array([0.0, 1.0, 0.0])
        mu = 1.0
        expected = jnp.array([0.0])
        result = adx.twobody.ivp.sigma_func(r, v, mu)
        assert jnp.allclose(result, expected)

    def test_nonzero_result(self) -> None:
        """Test with inputs that produce non-zero result."""
        r = jnp.array([1.0, 1.0, 0.0])
        v = jnp.array([1.0, 0.0, 0.0])
        mu = 1.0
        expected = jnp.array([1.0])
        result = adx.twobody.ivp.sigma_func(r, v, mu)
        assert jnp.allclose(result, expected)

    def test_array_inputs(self) -> None:
        """Test with array inputs."""
        r = jnp.array([[1.0, 0.0, 0.0], [2.0, 0.0, 0.0]])
        v = jnp.array([[0.0, 1.0, 0.0], [0.0, 2.0, 0.0]])
        mu = jnp.array([[1.0], [2.0]])
        expected = jnp.array([[0.0], [0.0]])
        result = adx.twobody.ivp.sigma_func(r, v, mu)
        assert jnp.allclose(result, expected)

    def test_broadcasting(self) -> None:
        """Test broadcasting capabilities."""
        r = jnp.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0]])
        v = jnp.array([1.0, 1.0, 0.0])
        mu = 4.0
        expected = jnp.array([[0.5], [1.0]])
        result = adx.twobody.ivp.sigma_func(r, v, mu)
        assert jnp.allclose(result, expected)

    def test_different_mu_values(self) -> None:
        """Test with different gravitational parameter values."""
        r = jnp.array([3.0, 0.0, 0.0])
        v = jnp.array([0.0, 2.0, 0.0])
        mu_values = jnp.array([1.0, 4.0, 9.0])
        expected = jnp.array([[0.0], [0.0], [0.0]])
        results = jnp.stack([adx.twobody.ivp.sigma_func(r, v, mu) for mu in mu_values])
        assert jnp.allclose(results, expected)



import jax.numpy as jnp
from jax.typing import ArrayLike
from jax import Array


def sigma_func(r: ArrayLike, v: ArrayLike, mu: ArrayLike) -> Array:
    r"""The sigma function

    Args:
        r: (...,3) The position vector.
        v: (...,3) The velocity vector.
        mu: The gravitational parameter.

    Returns:
        The value of the sigma function.

    Notes:
        The sigma function is defined as:
        $$
        \sigma = \frac{\boldsymbol{r} \cdot \boldsymbol{v}}{\sqrt{\mu}}
        $$
        where $\boldsymbol{r}$ is the position vector, $\boldsymbol{v}$ is the velocity vector, and $\mu$ is the gravitational parameter.

    References:
        Battin, 1999, pp.174.

    Examples:
        A simple example of calculating the sigma function with a position vector of [1, 0, 0], a velocity vector of [0, 1, 0], and a gravitational parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> r = jnp.array([1.0, 0.0, 0.0])
        >>> v = jnp.array([0.0, 1.0, 0.0])
        >>> mu = 1.0
        >>> adx.twobody.ivp.sigma_func(r, v, mu)
        Array([0.], dtype=float32)

        With broadcasting, you can calculate the sigma function for multiple position and velocity vectors:

        >>> r = jnp.array([[1.0, 0.0, 0.0], [2.0, 0.0, 0.0]])
        >>> v = jnp.array([[0.0, 1.0, 0.0], [0.0, 2.0, 0.0]])
        >>> mu = jnp.array([[1.0], [2.0]])
        >>> adx.twobody.ivp.sigma_func(r, v, mu)
        Array([[0.],
               [0.]], dtype=float32)
    """
    return jnp.sum(r * v, axis=-1, keepdims=True) / jnp.sqrt(mu)


def U0(chi: ArrayLike, alpha: ArrayLike) -> Array:
    r"""The universal function U0

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.


    Returns:
        The value of the universal function U0.

    Notes:
        The universal function U0 is defined as:
        $$
        U_0(\chi, \alpha) = \begin{cases}
        1 & \alpha = 0 \\
        \cos(\sqrt{\alpha} \chi) & \alpha > 0 \\
        \cosh(\sqrt{-\alpha} \chi) & \alpha < 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example of calculating the universal function U0 with a argument of 1.0 and a parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ivp.U0(chi, alpha)
        Array(0.5403..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the universal function U0 for multiple arguments and parameters:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = jnp.array([1.0, -2.0])
        >>> adx.twobody.ivp.U0(chi, alpha)
        Array([0.5403..., 8.4889...], dtype=float32)
    """
    conds = [
        alpha > 0,
        alpha < 0,
    ]
    choices = [
        jnp.cos(jnp.sqrt(alpha) * chi),
        jnp.cosh(jnp.sqrt(-alpha) * chi),
    ]
    return jnp.select(conds, choices, default=1.0)


def U1(chi: ArrayLike, alpha: ArrayLike) -> Array:
    r"""The universal function U1

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U1.

    Notes:
        The universal function U1 is defined as:
        $$
        U_1(\chi, \alpha) = \begin{cases}
        \chi & \alpha = 0 \\
        \frac{\sin(\sqrt{\alpha} \chi)}{\sqrt{\alpha}} & \alpha > 0 \\
        \frac{\sinh(\sqrt{-\alpha} \chi)}{\sqrt{-\alpha}} & \alpha < 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example of calculating the universal function U1 with a argument of 1.0 and a parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ivp.U1(chi, alpha)
        Array(0.8414..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the universal function U1 for multiple arguments and parameters:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = jnp.array([1.0, -2.0])
        >>> adx.twobody.ivp.U1(chi, alpha)
        Array([0.8414..., 5.9608...], dtype=float32)
    """
    conds = [
        alpha > 0,
        alpha < 0,
    ]
    choices = [
        jnp.sin(jnp.sqrt(alpha) * chi) / jnp.sqrt(alpha),
        jnp.sinh(jnp.sqrt(-alpha) * chi) / jnp.sqrt(-alpha),
    ]
    return jnp.select(conds, choices, default=chi)


def U2(chi: ArrayLike, alpha: ArrayLike) -> Array:
    r"""The universal function U2

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U2.

    Notes:
        The universal function U2 is defined as:
        $$
        U_2(\chi, \alpha) = \begin{cases}
        \frac{\chi^2}{2} & \alpha = 0 \\
        \frac{1 - \cos(\sqrt{\alpha} \chi)}{\alpha} & \alpha > 0 \\
        \frac{1 - \cosh(\sqrt{-\alpha} \chi)}{\alpha} & \alpha < 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example of calculating the universal function U2 with a argument of 1.0 and a parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ivp.U2(chi, alpha)
        Array(0.4596..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the universal function U2 for multiple arguments and parameters:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = jnp.array([1.0, -2.0])
        >>> adx.twobody.ivp.U2(chi, alpha)
        Array([0.4596..., 3.7444...], dtype=float32)
    """
    conds = [
        alpha > 0,
        alpha < 0,
    ]
    choices = [
        (1 - jnp.cos(jnp.sqrt(alpha) * chi)) / alpha,
        (1 - jnp.cosh(jnp.sqrt(-alpha) * chi)) / alpha,
    ]
    return jnp.select(conds, choices, default=chi**2 / 2.0)


def U3(chi: ArrayLike, alpha: ArrayLike) -> Array:
    r"""The universal function U3

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U3.

    Notes:
        The universal function U3 is defined as:
        $$
        U_3(\chi, \alpha) = \begin{cases}
        \frac{\chi^3}{6} & \alpha = 0 \\
        \frac{\sqrt{\alpha} \chi - \sin(\sqrt{\alpha} \chi)}{\alpha \sqrt{\alpha}} & \alpha > 0 \\
        \frac{\sqrt{-\alpha} \chi - \sinh(\sqrt{-\alpha} \chi)}{\alpha \sqrt{-\alpha}} & \alpha < 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.180.

    Examples:
        A simple example of calculating the universal function U3 with a argument of 1.0 and a parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ivp.U3(chi, alpha)
        Array(0.1585..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the universal function U3 for multiple arguments and parameters:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = jnp.array([1.0, -2.0])
        >>> adx.twobody.ivp.U3(chi, alpha)
        Array([0.1585..., 1.9804...], dtype=float32)
    """

    conds = [
        alpha > 0,
        alpha < 0,
    ]
    choices = [
        (jnp.sqrt(alpha) * chi - jnp.sin(jnp.sqrt(alpha) * chi))
        / alpha
        / jnp.sqrt(alpha),
        (jnp.sqrt(-alpha) * chi - jnp.sinh(jnp.sqrt(-alpha) * chi))
        / alpha
        / jnp.sqrt(-alpha),
    ]
    return jnp.select(conds, choices, default=chi**3 / 6.0)


def U4(chi: ArrayLike, alpha: ArrayLike) -> Array:
    r"""The universal function U4

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U4.

    Notes:
        The universal function U4 is defined as:
        $$
        U_4(\chi, \alpha) = \begin{cases}
        \frac{\chi^4}{24} & \alpha = 0 \\
        \frac{\chi^2 / 2 - U_2(\chi, \alpha)}{\alpha} & \alpha \neq 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.183.

    Examples:
        A simple example of calculating the universal function U4 with a argument of 1.0 and a parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ivp.U4(chi, alpha)
        Array(0.0403..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the universal function U4 for multiple arguments and parameters:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = jnp.array([1.0, -2.0])
        >>> adx.twobody.ivp.U4(chi, alpha)
        Array([0.0403..., 0.8722...], dtype=float32)
    """
    return jnp.where(
        jnp.isclose(alpha, 0.0), chi**4 / 24.0, (chi**2 / 2 - U2(chi, alpha)) / alpha
    )


def U5(chi: ArrayLike, alpha: ArrayLike) -> Array:
    r"""The universal function U5

    Args:
        chi: The generalized anomaly.
        alpha: The reciprocal of the semimajor axis.

    Returns:
        The value of the universal function U5.

    Notes:
        The universal function U5 is defined as:
        $$
        U_5(\chi, \alpha) = \begin{cases}
        \frac{\chi^5}{120} & \alpha = 0 \\
        \frac{\chi^3 / 6 - U_3(\chi, \alpha)}{\alpha} & \alpha \neq 0
        \end{cases}
        $$
        where $\chi$ is the generalized anomaly and $\alpha = \frac{1}{a}$ is the reciprocal of semimajor axis.

    References:
        Battin, 1999, pp.183.

    Examples:
        A simple example of calculating the universal function U5 with a argument of 1.0 and a parameter of 1.0:

        >>> import jax.numpy as jnp
        >>> import astrodynx as adx
        >>> chi = 1.0
        >>> alpha = 1.0
        >>> adx.twobody.ivp.U5(chi, alpha)
        Array(0.008137..., dtype=float32, weak_type=True)

        With broadcasting, you can calculate the universal function U5 for multiple arguments and parameters:

        >>> chi = jnp.array([1.0, 2.0])
        >>> alpha = jnp.array([1.0, -2.0])
        >>> adx.twobody.ivp.U5(chi, alpha)
        Array([0.008137..., 0.323536...], dtype=float32)
    """
    return jnp.where(
        jnp.isclose(alpha, 0.0), chi**5 / 120.0, (chi**3 / 6 - U3(chi, alpha)) / alpha
    )
